#include "rclcpp/rclcpp.hpp"
#include "sensor_msgs/msg/laser_scan.hpp"
#include "sensor_msgs/msg/point_cloud2.hpp"
#include "laser_geometry/laser_geometry.hpp"
#include "tf2_ros/buffer.h"
#include "tf2_ros/transform_listener.h"
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl_conversions/pcl_conversions.h>
#include <pcl/features/normal_3d.h>
#include <string>
#include <vector>
#include "nav2_costmap_2d/costmap_2d_ros.hpp"
#include "nav_msgs/msg/occupancy_grid.hpp"
#include "std_msgs/msg/header.hpp"
#include <math.h> 

class point_cloud_to_costmap : public rclcpp::Node{
public:
rclcpp::Publisher<nav_msgs::msg::OccupancyGrid>::SharedPtr publisher_;
point_cloud_to_costmap() : Node("rptc"), tf_buffer_(this->get_clock()), tf_listener_(tf_buffer_)
  {
    publisher_ = this->create_publisher<nav_msgs::msg::OccupancyGrid>("point_cloud_costmap", 10);

    auto topic_callback =
      [this](const sensor_msgs::msg::LaserScan::SharedPtr msg) -> void {
        RCLCPP_INFO(this->get_logger(), "I heard a LaserScan message");
        // Process the LaserScan message here

	      RCLCPP_INFO(this->get_logger(), "Angle min: %f", msg->angle_min);
        RCLCPP_INFO(this->get_logger(), "Angle max: %f", msg->angle_max);
        RCLCPP_INFO(this->get_logger(), "Angle increment: %f", msg->angle_increment);
        RCLCPP_INFO(this->get_logger(), "Time increment: %f", msg->time_increment);
        RCLCPP_INFO(this->get_logger(), "Scan time: %f", msg->scan_time);
        RCLCPP_INFO(this->get_logger(), "Range min: %f", msg->range_min);
        RCLCPP_INFO(this->get_logger(), "Range max: %f", msg->range_max);
        RCLCPP_INFO(this->get_logger(), "Ranges size: %zu", msg->ranges.size());
        // Print out the first few range values as an example
        for (size_t i = 0; i < std::min(msg->ranges.size(), static_cast<size_t>(5)); ++i) {
          RCLCPP_INFO(this->get_logger(), "Range[%zu]: %f", i, msg->ranges[i]);
        }

        nav_msgs::msg::OccupancyGrid costmap = process_local_scan(msg);
        if(costmap.header.frame_id != "invalid"){
          publisher_->publish(costmap);
        }
      };
    subscription_ =
      this->create_subscription<sensor_msgs::msg::LaserScan>("/lidar", 10, topic_callback);
  }

private:
  rclcpp::Subscription<sensor_msgs::msg::LaserScan>::SharedPtr subscription_;
  laser_geometry::LaserProjection projector_;
  tf2_ros::Buffer tf_buffer_;
  tf2_ros::TransformListener tf_listener_;  

  typedef pcl::PointXYZ PointT;
  typedef pcl::PointCloud<PointT> PointCloudT;
  typedef pcl::PointNormal PointNT;
  typedef pcl::PointCloud<PointNT> PointCloudNT;

  nav_msgs::msg::OccupancyGrid process_local_scan(const sensor_msgs::msg::LaserScan::SharedPtr msg){
    
    // Turn scan msg into pointcloud msg
    sensor_msgs::msg::PointCloud2::SharedPtr cloud_msg = std::make_shared<sensor_msgs::msg::PointCloud2>();
    try {
      projector_.transformLaserScanToPointCloud("vehicle_blue/chassis/gpu_lidar", *msg, *cloud_msg, tf_buffer_);
      RCLCPP_INFO(this->get_logger(), "Converted LaserScan to PointCloud2");
    } catch (const std::exception& e) {
      RCLCPP_ERROR(this->get_logger(), "Error converting LaserScan to PointCloud2: %s", e.what());
    }

    // Check if the PointCloud2 message is empty
    if (cloud_msg->data.empty()) {
      RCLCPP_ERROR(this->get_logger(), "PointCloud2 message is empty");

      nav_msgs::msg::OccupancyGrid costmap = init_costmap(0.0f, 0, 0);
      costmap.header.frame_id = "invalid";
      return costmap;
    }

    // Turn pointcloud msg into pcl pointcloud
    PointCloudT::Ptr cloud(new PointCloudT);
    PointCloudNT::Ptr cloud_with_normals(new PointCloudNT);
    pcl::fromROSMsg(*cloud_msg, *cloud);

    pcl::NormalEstimation<PointT, PointNT> ne;
    ne.setInputCloud(cloud);

    pcl::search::KdTree<PointT>::Ptr tree(new pcl::search::KdTree<PointT>());
    ne.setSearchMethod(tree);

    ne.setRadiusSearch(0.03);

    ne.compute(*cloud_with_normals);

    for (const auto& point : cloud_with_normals->points) {
      RCLCPP_INFO(this->get_logger(), "Point: (%.2f, %.2f, %.2f), Normal: (%.2f, %.2f, %.2f)",
                  point.x, point.y, point.z, point.normal_x, point.normal_y, point.normal_z);
    }

    // Create costmap with 5x5 meters 100x100 cells
    float resolution = 0.05f; // res in meters
    int dimension = 100; // grid cells
    int dim_half = dimension / 2;

    // Create 2d grid for gradients
    std::vector<std::vector<PointNT>> arr(dimension + 2, std::vector<PointNT>(dimension + 2, PointNT()));

    // Sum up
    for (const auto& point : cloud_with_normals->points) {
      // Get X and Y coordinate in grid
      int x_coord = static_cast<int>(point.x / resolution) + dim_half;
      int y_coord = static_cast<int>(point.y / resolution) + dim_half;

      // Ensure the coordinates are within bounds
      if (x_coord >= 0 && x_coord < dimension + 2 && y_coord >= 0 && y_coord < dimension + 2) {
        arr[x_coord][y_coord].normal_x += point.normal_x;
        arr[x_coord][y_coord].normal_y += point.normal_y;
        arr[x_coord][y_coord].normal_z += point.normal_z;
      }
    }

    // Normalize
    for (int x = 0; x < dimension + 2; ++x) {
      for (int y = 0; y < dimension + 2; ++y) {
        PointNT& point = arr[x][y];
        float magnitude = std::sqrt(point.normal_x * point.normal_x +
                                    point.normal_y * point.normal_y +
                                    point.normal_z * point.normal_z);
        if (magnitude > 0) {
          point.normal_x /= magnitude;
          point.normal_y /= magnitude;
          point.normal_z /= magnitude;
        }
      }
    }

    // Calculate gradients between cells and
    // Transfer to costmap
    nav_msgs::msg::OccupancyGrid costmap = init_costmap(resolution, dimension, dimension);
    for (int x = 0; x < dimension; ++x) {
      for (int y = 0; y < dimension; ++y) {

        float score = 0.0f;
        PointNT self = arr[x+1][y+1];
        for(int i = 0; i <= 2; ++i){
          for(int j = 0; j <= 2; ++j){
            if(i == 1 && j == 1){
              continue;
            }
            score += dot_product(self, arr[x+i][y+j]);
          }
        }

        //PointNT neighbour_tl = arr[x][y];
        //PointNT neighbour_tt = arr[x+1][y];
        //PointNT neighbour_tr = arr[x+2][y];

        //PointNT neighbour_ml = arr[x][y+1];
        //PointNT neighbour_mr = arr[x+2][y+1];

        //PointNT neighbour_bl = arr[x][y+2];
        //PointNT neighbour_bb = arr[x+1][y+2];
        //PointNT neighbour_br = arr[x+2][y+2];

        //PointNT self = arr[x+1][y+1];

        //float score = dot_product(self, neighbour_tl) + dot_product(self, neighbour_tt)

        costmap.data[x + (y * dimension)] = score;
      }
    }
    return costmap;
  }

  nav_msgs::msg::OccupancyGrid init_costmap(float resolution_in_meters, int width, int height) {
    nav_msgs::msg::OccupancyGrid costmap;
  
    // Set the header
    costmap.header.stamp = rclcpp::Clock().now();
    costmap.header.frame_id = "map";
  
    // Define the costmap properties
    costmap.info.resolution = resolution_in_meters; // Resolution in meters per cell
    costmap.info.width = width;       // Width in cells
    costmap.info.height = height;      // Height in cells
    costmap.info.origin.position.x = 0.0; // Origin of the costmap
    costmap.info.origin.position.y = 0.0;
    costmap.info.origin.position.z = 0.0;
    costmap.info.origin.orientation.w = 1.0;
  
    // Initialize the costmap data with unknown values (-1)
    costmap.data.resize(costmap.info.width * costmap.info.height, -1);
  
    return costmap;
  }

  double dot_product(const PointNT& p1, const PointNT& p2) {
    return p1.normal_x * p2.normal_x + p1.normal_y * p2.normal_y + p1.normal_z * p2.normal_z;
  }
};

int main(int argc, char ** argv)
{
  printf("start robinspec pointcloud to costmap\n");
  rclcpp::init(argc, argv);
  auto node = std::make_shared<point_cloud_to_costmap>();
  rclcpp::spin(node);
  rclcpp::shutdown();
  printf("exit\n");
  return 0;
}
